################################################################################
#
#DATANAME JEM-X 2 - SKY - IMAGE
#
#DESCRIPTION
# Contains the Jem-X 2 sky image obtained for a single Science Window.
#
#TEMPLATE JMX2-SKY.-IMA
#
#CONTENT IMAGE
#DATADEF : Image pixel value
#AXDEF 1 : Phi angle (CAR) or X coordinate (TAN)
#AXDEF 2 : Theta angle (CAR) or Y coordinate (TAN)
#
#REMARKS
# This data structure can store either a cartesian (CAR) or a tangential (TAN)
# projection of the celestial sphere. The coordinate system is specific to this
# data structure. The standard keywords defined in this data structure provide
# all information require to compute RA and DEC (or GLON GLAT, or ELON ELAT)
# of any pixel.
#.
# This data structure has been defined according to the standard representations
# of celestial coordinates in FITS proposed by Greisen and Calabretta (AA,
# preprint)
#.
# The IMATYPE keyword can have the following values:
# INTENSITY, VARIANCE, RESIDUAL or SIGNIFICANCE
#.
# The UTC format is YYYY-MM-DDTHH:MM:SS
#
#RP obs/scw
#DB No
#CC Yes
#
#CHANGES
# 1.6.1 : Template creation
# 2.0.1 : Renamed template (previously JMX2-INTE-IMA)
#       : Added ISDCLEVL keyword
#       : Added following keywords: BUNIT, BSCALE, BZERO, CTYPE1, CRPIX1,
#       : CRVAL1, CDELT1, CUNIT1, CTYPE2, CRPIX2, CRVAL2, CDELT2, CUNIT2,
#       : CROTA2, PC001001, PC002002, PC001002, PC002001, LONGPOLE, LATPOLE,
#       : RADECSYS, EQUINOX, HDUCLASS, HDUDOC, HDUVERS, HDUCLAS1,
#       : E_MIN, E_MEAN, E_MAX, IMATYPE.
#       : Changed NAXIS1 and NAXIS2 value to Integer (previously 200)
#       : BSCALE set to 1, BZERO set to 0, CUNIT1 and CUNIT2 set to deg
# 2.0.2 : Removed the four PC00i00i (i=1,2) keywords
# 2.0.3 : Removed the LONGPOLE and LATPOLE keywords
# 2.1   : Set BUNIT value to count/s/cm**2
# 2.1.3 : Replaced CDELT1,CDELT2 and CROTA2 keywords
#       : by the CD1_1, CD1_2, CD2_1 and CD2_2 keywords
#       : Renamed DATE-END keyword (previously DATE_END) 
# 2.2   : Removed EXPID keyword
#       : Added ONTIME, DEADC and EXPOSURE keywords
#       : Changed HDUVERS value from 1.0 to 3.0
#       : Put back the LONGPOLE and LATPOLE keywords
# 2.3   : Added SWBOUND keyword
# 3.4   : Replaced ROWIDEN1 and ROWIDEN2 keywords by SHDIDEN
#       : Removed DATE-OBS and DATE-END keywords
#       : Template put under Configuration Control
# 5.5   : Changed BUNIT value from count/s/cm**2 to count/s
#
################################################################################
XTENSION	IMAGE		/ Image extension
BITPIX		-32		/ IEEE 32-bit floating point values
NAXIS		2		/ Number of data axes
NAXIS1		Integer		/ Length of data axis 1
NAXIS2		Integer		/ Length of data axis 2
EXTNAME		JMX2-SKY.-IMA	/ Extension name
EXTREL		'5.5'		/ ISDC release number
BASETYPE	DAL_ARRAY	/ Data Access Layer base type
TELESCOP	INTEGRAL	/ Telescope or mission name
ORIGIN		ISDC		/ Origin of FITS file
INSTRUME	JMX2		/ Instrument name
ISDCLEVL	IMA		/ ISDC level of data processing
CREATOR		String		/ Program that created this FITS file
CONFIGUR	String		/ Software configuration
DATE		UTC_format	/ FITS file creation date
MJDREF		51544.0		/ Modified Julian Date of time origin
TIMESYS		TT		/ Time frame system
TIMEUNIT	d		/ Time unit
TIMEREF		LOCAL		/ Time reference frame
REVOL		Integer		/ Revolution number
SWID		String		/ Science Window identifier
SW_TYPE		String		/ Type of the Science Window
SWBOUND		String		/ Reason for Science Window ending
OBTSTART	OBT_format	/ OBT of the start of the Science Window
OBTEND		OBT_format	/ OBT of the end of the Science Window
TSTART		Real		/ Start time of the observation
TSTOP		Real		/ End time of the observation
TELAPSE		Real		/ [s] Total observation elapsed time
ONTIME		Real		/ Mean ontime for all live detectors
DEADC		Real		/ Mean deadtime correction for all live detectors
EXPOSURE	Real		/ Mean exposure time over the detector plane
BSCALE		1		/ Real value=value*BSCALE + BZERO
BZERO		0		/ Offset applied to true pixel values
BUNIT		count/s		/ Unit for pixel values
CTYPE1		String		/ LONGPROJ where LONG can be RA, GLON, ELON and PROJ can be CAR, TAN or AIT
CRPIX1		Real		/ Pixel at reference point
CRVAL1		Real		/ LONG at the reference value
CUNIT1		deg		/ Physical units of axis 1
CTYPE2		String		/ LAT-PROJ where LAT can be DEC, GLAT, ELAT and PROJ can be CAR, TAN or AIT
CRPIX2		Real		/ Pixel at reference point
CRVAL2		Real		/ LAT at the reference value
CUNIT2		deg		/ Physical units of axis 2
CD1_1		1.0		/ Element (1,1) of coordinate transf. matrix (default 1)
CD1_2		0.0		/ Element (1,2) of coordinate transf. matrix (default 0)
CD2_1		0.0		/ Element (2,1) of coordinate transf. matrix (default 0)
CD2_2		1.0		/ Element (2,2) of coordinate transf. matrix (default 1)
LONGPOLE	180		/ Longitude in the native coord. system of the std system's N pole
LATPOLE		0		/ Latitude in the native coord. system of the std system's N pole
RADECSYS	FK5		/ Stellar reference frame
EQUINOX		2000.0		/ Coordinate system equinox
HDUCLASS	OGIP		/ Format conforms mostly to OGIP standards
HDUDOC		'ISDC-JEMX ADD'	/ Document in which the format is defined
HDUVERS		'3.0'		/ Version of format
HDUCLAS1	IMAGE		/ Sky Image
CHANTYPE	PI		/ Type of detector channels
CHANMIN		Integer		/ Lowest channel of the energy range
CHANMAX		Integer		/ Highest channel of the energy range
SHDIDEN		Integer		/ Shadowgram identification number
EVNTYPES	Integer		/ Code for event types (0:FULL, 1:REST, -1:both)
E_MIN		Real		/ [keV] Lower energy limit
E_MAX		Real		/ [keV] Upper energy limit
E_MEAN		Real		/ [keV] Mean energy
IMATYPE		String		/ Type of image
