/*
Routine to compute mosaic weigthed sky images. Jrme Chenevez, DSRI.
Version 0  Sept 2003 : Experimental version (memory problems solved by declaration of big arrays outside the main pgm).
Version 1  Oct  2003 : First release (the user has to set input images dimensions "by hand" - solved in next versions).
Version 2  Dec  2003 : Use of a parameter file. New weighting method by NJW. Introduction of larger input images from CBJ.
Version 3  Feb  2004 : Improved selection capabilities for NL input images. Ranking of weights for a given sky position.
Version 4  Apr  2004 : Advanced I/O formats following agreements described in JEM-X TN-14.
Version 5  Aug  2004 : Implementation of NJW's subroutine to compute variance maps. The keyword VIGNDOL is replaced by an environment variable.
*/

#include<stdio.h>
#include<math.h>
#include<stdlib.h>
#include<string.h>
#include<jemx.h>
#include<isdc.h>
#include<dal3aux.h>

#define NB_IMAGES 999       /* Maximum number of input images here (may be increased) */
#define XD 825              /* X dim. of input images  (NJW: 200, or up to 825 ; NL: 255) */
#define YD 825              /* Y dim. of input images  (CBJ: New:525, Old:262) */
#define MOSAICSIZE 1000000

/* Declare big arrays as global variables ! */
  int mark[MOSAICSIZE];
  double pval_o[MOSAICSIZE], var_o[MOSAICSIZE], exp_o[MOSAICSIZE], var_s[MOSAICSIZE], exp_s[MOSAICSIZE], rat_s[MOSAICSIZE], t_exp[MOSAICSIZE], v_exp[MOSAICSIZE], weight[MOSAICSIZE], weigh2[MOSAICSIZE];
  double pvalue[XD*YD], variance[XD*YD];
  double pval_i[XD][YD], var_i[XD][YD], exp_i[XD][YD];
  double vign[XD][YD], vigntmp[XD][YD];
  double vign1D[XD*YD];
  double pvalcor_f[525*525];  /* CBJ's new correction array */
  float value_in[XD*YD], value_out[XD*YD];

/* ********************************************************** */
#define      sky_xdim     255
#define      sky_ydim     255
#define      skydim       sky_xdim*sky_ydim

  void clean(float sky_r[skydim], float skyim[skydim], int bol);

#define PI  3.14159265

  float circ_area1( float a, float b, float c );
  float thruput_instr( float a, float b );
  int j_vign_matrix( int sky_dimx, int sky_dimy, float cdelt1, float cdelt2, double roll_a, double *vign_matrix, int _status );

/* ********************************************************** */

main (int argc, char *argv[])
{
  int i=0, j=0, k=0, loop=0, overloop=0, cnt=0;
  int loopM[NB_IMAGES], rank[NB_IMAGES], testM=0;
  int xdim=XD, ydim=YD, dimx=(int)(sqrt(MOSAICSIZE)), dimy=(int)(sqrt(MOSAICSIZE));
  int numHDU=2, numAxes=2;
  int p=0, q=0, pix[2]={0, 0};
  int index=0, index1=0, index2=0;
  int status=ISDC_OK, status1=ISDC_OK;
  int isub, nsub, psub=0, qsub=0;
  int ext[NB_IMAGES];
  int maxImages=-1, numEntries=0, numImages = 0;  /* Number of images found  */
  int pvalcor[262*262];  /* CBJ's correction array */
  int nex=1;
  int useA=0, suffix=0;
  int useE=0, ecode_min=0, ecode_max=0, e_dol=2;
  int plot_T=0, plot_I=0, plot_E=0, plot_V=0, plot_W=0, plot_W2=0, plot_SoE=0, plot_RVE=0;
  long startValues[2] = {1L, 1L};
  long endValues[2] = {(long)(XD), (long)(YD)};
  long numValues=0L;
  long ecode[NB_IMAGES];
  double pi, dr;
  float d_a=0., d_b=0., d_c=0., d_D=0.;
  float wei=1., rwei[NB_IMAGES], rweiMax=0.;
  float dsub, reduc, dp=0., dq=0.;
  float cdelta = 0.016667, cdeltb = 0.06667;
  double ra_o=0., de_o=0., ra_x=0., de_x=0.;
  double t1[3][3], t2[3][3], t[3][3];
  double t1_o[3][3], t2_o[3][3], t_o[3][3], o_t[3][3];
  double NaN = NAN;
  double ra=0., dec=0.;
  double centerA=-1.0, centerD=-1.0;
  double radius=5., diamout=15.;
  double exp_t[NB_IMAGES];
  double crpix1[NB_IMAGES], crpix2[NB_IMAGES];
  double cd1_1[NB_IMAGES], cd2_1[NB_IMAGES],cd1_2[NB_IMAGES], cd2_2[NB_IMAGES];
  double crval1[NB_IMAGES], crval2[NB_IMAGES];
  double crval1_min=0., crval1_max=0., crval2_min=0., crval2_max=0., crval1_o=0., crval2_o=0.;
  double crpix1_o=0., crpix2_o=0.;
  double cdelt=0.03, cd1_1_o=0., cd1_2_o=0., cd2_1_o=0., cd2_2_o=0.;
  double yz[2], ayz[2], vi[3], ve[3], vi_o[3];
  double roll=0., maxvign=0.;
  double emin=0., emax=0., min_e=80., max_e=0.;
  double e_min[NB_IMAGES], e_max[NB_IMAGES];
  char DOLstring[NB_IMAGES][99], DOLstring_1[NB_IMAGES][99], DOLstring_2[99];  /* DAL_FILENAME_STRING */
  char VIGNDOL[99] = "VIGN_DIR/NL_VIGNET_1.fits\[1]";
  char vigndol[99], vign_f[99], vign_s[99], vign_d[99];
  char vignef[] = "VIGNET";
  char fitstring[99], skymaps[99], mosaic[99];
  char filexpstring[NB_IMAGES][99];
  char IMACASE[] = "'BASIC_01'", fine_str[] = "'FINE_01 '", basic_str[] = "'BASIC_01'";
  char template[] = "JMX-SKY.-IMA-IDX.tpl";
/*  char plot_T='N', plot_I='N', plot_E='N', plot_V='N', plot_W='N', plot_W2='N', plot_SoE='N', plot_RVE='N'; */
  void *arrayPtr = NULL;
  void *arrayBuff = NULL;
  FILE *vignet;
  FILE *fp1;

  dal_dataType DALtype = DAL_DOUBLE;
  dal_element *dolPtr = NULL;
  dal_element *skyPtr = NULL;
  dal_element *pvlPtr = NULL;
  dal_element *expPtr = NULL;
  dal_element *varPtr = NULL;
  dal_element *sw_Ptr = NULL;
  dal_element *sw2Ptr = NULL;
  dal_element *mapPtr = NULL;
  dal_element *sqtPtr = NULL;
  dal_element *timPtr = NULL;
  dal_element *texPtr = NULL;
  dal_element *extPtr = NULL;
  fitsfile *fitsinPtr = NULL;

/* Statements begin */
  pi = acos (-1.);
  dr = pi/180.;
  isub = 2;
  nsub = 2*isub + 1;
  dsub = 1./nsub;
  reduc = dsub*dsub;

/* Read input parameters */
  status = PILInit(argc, argv);
  if (status != ISDC_OK) {
    printf("%d: PIL initialisation error\n", status);
    return status;
  }
  status = PILGetString("fileList", skymaps);
  if (status != ISDC_OK) {
    printf("%d: Could not read fileList parameter\n", status);
    return (status);
  }
  status = PILGetReal("radiuSelect", &radius);
  if (status != ISDC_OK) {
    printf("Could not read radiuSelect parameter\n");
    return (status);
  }
  status = PILGetReal("diameter", &diamout);
  if (status != ISDC_OK) {
    printf("Could not read diameter parameter\n");
    return (status);
  }
  status = PILGetReal("cdelt", &cdelt);
  if (status != ISDC_OK) {
    printf("Could not read cdelt parameter\n");
    return (status);
  }
  status = PILGetReal("RAcenter", &centerA);
  if (status != ISDC_OK) {
    printf("Could not read RAcenter parameter\n");
    return (status);
  }
  status = PILGetReal("DECcenter", &centerD);
  if (status != ISDC_OK) {
    printf("Could not read DECcenter parameter\n");
    return (status);
  }
  status = PILGetInt("inputCode", &useA);
  if (status != ISDC_OK) {
    printf("Could not read inputCode parameter\n");
    return (status);
  }
  status = PILGetReal("eminSelect", &emin);
  if (status != ISDC_OK) {
    printf("Could not read eminSelect parameter\n");
    return (status);
  }
  status = PILGetReal("emaxSelect", &emax);
  if (status != ISDC_OK) {
    printf("Could not read emaxSelect parameter\n");
    return (status);
  }
  status = PILGetBool("useEcode", &useE);
  if (status != ISDC_OK) {
    printf("Could not read useEcode parameter\n");
    return (status);
  }
  status = PILGetInt("ecodeMin", &ecode_min);
  if (status != ISDC_OK) {
    printf("Could not read ecodeMin parameter\n");
    return (status);
  }
  status = PILGetInt("ecodeMax", &ecode_max);
  if (status != ISDC_OK) {
    printf("Could not read ecodeMax parameter\n");
    return (status);
  }
  status = PILGetString("outfile", mosaic);
  if (status != ISDC_OK) {
    printf("Could not read outfile parameter\n");
    return (status);
  }
  status = PILGetBool("outSuffix", &suffix);
  if (status != ISDC_OK) {
    printf("Could not read outSuffix parameter\n");
    return (status);
  }
  status = PILGetBool("viewTime", &plot_T);
  if (status != ISDC_OK) {
    printf("Could not read viewTime parameter\n");
    return (status);
  }
  status = PILGetBool("viewIntens", &plot_I);
  if (status != ISDC_OK) {
    printf("Could not read viewIntens parameter\n");
    return (status);
  }
  status = PILGetBool("viewExpo", &plot_E);
  if (status != ISDC_OK) {
    printf("Could not read viewExpo parameter\n");
    return (status);
  }
  status = PILGetBool("viewVar", &plot_V);
  if (status != ISDC_OK) {
    printf("Could not read viewVar parameter\n");
    return (status);
  }
  status = PILGetBool("viewWeight", &plot_W);
  if (status != ISDC_OK) {
    printf("Could not read viewWeight parameter\n");
    return (status);
  }
  status = PILGetBool("viewWeig2", &plot_W2);
  if (status != ISDC_OK) {
    printf("Could not read viewWeig2 parameter\n");
    return (status);
  }
  status = PILGetBool("viewSignal", &plot_SoE);
  if (status != ISDC_OK) {
    printf("Could not read viewSignal parameter\n");
    return (status);
  }
  status = PILGetBool("viewNoise", &plot_RVE);
  if (status != ISDC_OK) {
    printf("Could not read viewNoise parameter\n");
    return (status);
  }
  status = PILGetInt("maxNumFiles", &maxImages);
  if (status != ISDC_OK) {
    printf("Could not read maxNumFiles parameter\n");
    return (status);
  }
  status = PILGetString("vignDir", vign_d);
  if (status != ISDC_OK) {
    printf("Could not read vignDir parameter\n");
    return (status);
  }
  strcat(vign_d, "/");
  PILClose(status);

/* Look for number and size of input images */
  if( (fp1 = fopen(skymaps, "rt")) == NULL ) {
     printf("Could not open skymaps list file\n");
     return;
  }

  while (fscanf(fp1, "%s", fitstring) != EOF){
    numEntries++;
    if (fitstring[0] == 35) continue;
    numImages++;
  }
  if (fitstring[0] == 35){
    printf("*** The last file in the input list must not be commented out... Sorry, you have to remove it! ***\n");
    return;
  }

  if (useA <=0) strcat(fitstring, "[2]");  /* First skymap extension for NJW and CBJ */
  else  strcat(fitstring, "[1]");          /* NL's first extension is 1 */
  fits_open_file(&fitsinPtr, fitstring, READONLY, &status); /* Open input fits file to get Ptr */
  fits_read_key_lng(fitsinPtr, "NAXIS1", &numValues, NULL, &status); /* Get images dimensions in the input fits */
/*fits_read_keyword(fitsinPtr, "VIGNDOL", vigndol, NULL, &status);  This may be commented out in a further version */
  if (useA < 0) {
    fits_read_keyword(fitsinPtr, "IMACASE", IMACASE, NULL, &status);
    fits_read_key_dbl(fitsinPtr, "ROLL", &roll, NULL, &status);
  }
  fits_close_file(fitsinPtr, &status); /* Close input fits file */
  rewind(fp1);

  if (numImages > NB_IMAGES){
    printf("There are more than %d images! Please set the value NB_IMAGES > %d and recompile the program\n", NB_IMAGES, numImages);
    return;
  }
  else printf("There are %d input images of dimensions %ld * %ld pixels\n", numImages, numValues, numValues);
  if ((maxImages > 0) && (numImages > maxImages)){
    numImages = maxImages;
    printf("Only the first %d images will be used\n", maxImages);
  }

/* Initialisations */

  ydim = xdim = (int)(numValues);   /* Assuming that the input skymaps have always square dimensions */
  endValues[0] = numValues;
  endValues[1] = numValues;

  for (i=0; i<numImages; i++){
    loopM[i] = NB_IMAGES;
    rank[i] = 0;
    rwei[i] = -1.;
    ext[i] = 0;
    crpix1[i] = crpix2[i] = 0;
    exp_t[i] = 0.;
    cd1_1[i] = cd2_1[i] = cd1_2[i] = cd2_2[i] = 0.;
    crval1[i] = crval2[i] = 0.;
    e_min[i] = e_max[i] = 0.;
    ecode[i] = 0L;
  }
  for (i=0; i<MOSAICSIZE; i++){
    pval_o[i] = 0.;
    var_o[i] = 0.;
    exp_o[i] = 0.;
    var_s[i] = 0.;
    exp_s[i] = 0.;;
    rat_s[i] = 0.;
    t_exp[i] = 0.;
    v_exp[i] = 0.;
    weight[i] = 0.;
    weigh2[i] = 0.;
  }
  for (i=0; i<xdim; i++)
    for (j=0; j<ydim; j++){
      pval_i[i][j] = NaN;
      var_i[i][j] = 0.;
      exp_i[i][j] = 0.;
      vign[i][j] = 0.;
    }
  for (i=0; i<xdim*ydim; i++){
    pvalue[i] = NaN;
    variance[i] = 0.;
    vign1D[i] = 0.;
  }
  for (i=0; i<3; i++){
    vi_o[i] = 0.;
    ve[i] = 0.;
    for (j=0; j<3; j++){
      t1[i][j] = t2[i][j] = t[i][j] =0.;
      t1_o[i][j] = t2_o[i][j] = t_o[i][j] = 0.;
      o_t[i][j] = 0.;
    }
  }

/* Initialisations done */

/* Read vignetting-function array (in mm2) from txt file (NL or CBJ) 
  if (useA > 0) strcpy(vign_f, "illum_l_1.txt");
  else if (useA == 0){
    if (xdim == 262){
      strcpy(vign_s, vign_d);
      strcat(vign_s, "cbj_vig.dat");
      if( (vignet = fopen(vign_s,"rt")) == NULL ) {
        printf("Could not open the file named: %s\n", vign_s);
        return(0);
      }
      j = 0;
      while (j<xdim*ydim){
        for (i=0; i<43; i++){
          fgets(vign_s, 99, vignet);
          sscanf(vign_s, "%d %d %d %d %d %d", &pvalcor[j++],&pvalcor[j++],&pvalcor[j++],&pvalcor[j++],&pvalcor[j++],&pvalcor[j++]);
        }
        fgets(vign_s, 99, vignet);
        sscanf(vign_s, "%d %d %d %d", &pvalcor[j++],&pvalcor[j++],&pvalcor[j++],&pvalcor[j++]);
      }
      fclose(vignet);
      strcpy(vign_f, "vignet_262_2+.txt");
    }
    else if (xdim == 525){
      strcpy(vign_s, vign_d);
      strcat(vign_s, "bckim_1_5.txt");
      if( (vignet = fopen(vign_s,"rt")) == NULL ) {
        printf("Could not open the file named: %s\n", vign_s);
        return(0);
      }
      j = 0;
      while (j<xdim*ydim){
        for (i=0; i<131; i++){
          fgets(vign_s, 99, vignet);
          sscanf(vign_s, "%lf %lf %f %lf", &pvalcor_f[j++],&pvalcor_f[j++],&pvalcor_f[j++],&pvalcor_f[j++]);
        }
        fgets(vign_s, 99, vignet);
        sscanf(vign_s, "%lf", &pvalcor_f[j++]);
      }
      fclose(vignet);
      strcpy(vign_f, "vignet_525_2+.txt");
    }
  }
  strcat(vign_d, vign_f);
  printf("Use of %s for vignetting array\n", vign_d);
  if( (vignet = fopen(vign_d,"rt")) == NULL ) {
    printf("Could not open %s\n", vign_d);
    return(0);
  }
  for (i=0; i<xdim; i++)
    for (j=0; j<ydim; j++)
      if(1 !=  fscanf(vignet,"%lf", &vign[i][j])){
        fprintf(stderr,"Can't read vignetting at [%d,%d]\n", i, j);
        return 1;
      }
  fclose(vignet);
  for (i=1; i<xdim-1; i++)
    for (j=1; j<ydim-1; j++){
      vigntmp[i][j] = (vign[i-1][j-1]+vign[i][j-1]+vign[i+1][j-1]+\
      vign[i+1][j]  +vign[i][j]  +vign[i-1][j]+\
      vign[i-1][j+1]+vign[i][j+1]+vign[i+1][j+1])/9.;
      if (vigntmp[i][j] > maxvign) maxvign = vigntmp[i][j];
    }
  printf("Smoothed MaxVign = %f\n", maxvign);
  for (i=0; i<xdim; i++)
    for (j=0; j<ydim; j++){
      vign[i][j] = vigntmp[i][j]/maxvign;
      index = j*xdim + i;
      vign1D[index] = vign[i][j];
    }
*/
/* Convert vignetting array to FITS format 
  status = DALobjectCreate(vignef, DAL_DISK, "JMX-SKY.-IMA.tpl", &skyPtr, status);
  if (status != ISDC_OK){
    printf("Cannot create vignet object, status = %d\n",status);
    return status;
  }
  status = DALarrayModStruct(skyPtr, DALtype, numAxes, endValues, status);
  if( status != ISDC_OK ) printf("DALarrayModStruct status : %d\n", status);
  status = DALattributePutChar(skyPtr, "IMATYPE", "VIGNETTING", " ", "The image shows vignetting function", status);
  arrayBuff = vign1D;
  status = DALarrayPutSection(skyPtr, numAxes, startValues, endValues, DALtype, arrayBuff, status);
  if (status != ISDC_OK){
    printf("Cannot put section vign object, status = %d\n",status);
    return status;
  }
  status = DALobjectClose(skyPtr, DAL_SAVE, status);
  if (status != ISDC_OK){
    printf("Cannot close object vignef! status = %d\n",status);
    return status;
  }
*/
/* Vignetting function (normalised to 1) computed (NJW) or read from FITS file */
    if (useA < 0) {
      arrayPtr = vign1D;
      if (strncmp(IMACASE, fine_str, 5) == 0) {
        status = j_vign_matrix(xdim, ydim, cdelta, cdelta, 0.0, arrayPtr, status);
        printf("Use of auto_computed Fine imaging Vignetting function\n");
      }
      else if (strncmp(IMACASE, basic_str, 5) == 0) {
        status = j_vign_matrix(xdim, ydim, cdeltb, cdeltb, roll, arrayPtr, status);
        printf("Use of auto-computed Basic imaging Vignetting function\n");
      }
      else printf("*** No IMACASE defined! ***\n");
    }
    else {
      if (getenv ("VIGNDOL")) strcpy(vign_f, getenv ("VIGNDOL")); /* This may not be ISDC compliant */
      else if (strlen(vigndol) != 0) strcat(vign_f, vigndol);
      else {
        if (useA == 0) strcpy(vigndol, "CBJ_VIGNET.fits[1]");  /* The vigndol should be given either by an environment variable */
        if (useA > 0) strcpy(vigndol, "NL_VIGNET.fits[1]");    /* or by a keyword, but a fixed vigndol is used here. */
        strcpy(vign_f, vign_d);
        strcat(vign_f, vigndol);
      }
      printf("Use of %s for vignetting function\n", vign_f);
      status = DALobjectOpen(vign_f, &skyPtr, status);
      if (status != ISDC_OK){
        printf("Cannot open file %s !!!\n", vign_f);
        return status;
      }
      arrayPtr = vign1D;
      status = DALarrayGetSection(skyPtr, numAxes, startValues, endValues, &DALtype, &numValues, arrayPtr, status);
      if (status != ISDC_OK){
        printf("Cannot get section vign1D from %s\n", vign_f);
        return status;
      }
      status = DALobjectClose(skyPtr, DAL_SAVE, status);
      if (status != ISDC_OK){
        printf("Cannot close %s\n", vign_f);
        return status;
      }
    }
    for (i=0; i<xdim; i++) 
      for (j=0; j<ydim; j++){
        index = j*xdim + i;
        vign[i][j] = vign1D[index];
      }

/* Read HEADER attributes of each input images */
  cnt = 0;
  for (overloop=0; overloop<numEntries; overloop++){  /* Loop over input skymaps */
    if(1 != fscanf(fp1, "%s", fitstring)){
      fprintf(stderr, "Can't read fits file name in skymaps list file\n");
      return 1;
    }
    if (fitstring[0] == 35){
      cnt++;
      continue;
    }
    loop = overloop-cnt;
    strcpy(DOLstring[loop],fitstring);
    if (strcmp(DOLstring[loop], DOLstring[loop-1]) == 0){
      printf("*** Stop: DOL %s is repeated! ***\n", DOLstring[loop]);
      return 1;
    }
    fits_open_file(&fitsinPtr, DOLstring[loop], READONLY, &status); /* Open input fits file to get Ptr */
    fits_get_num_hdus(fitsinPtr, &numHDU, &status); /* Get total number of extensions in the input fits */
    fits_close_file(fitsinPtr, &status); /* Close input fits file */
    strcpy(DOLstring_1[loop], DOLstring[loop]);

    if (useA < 0) sprintf(DOLstring_1[loop], "%s[%d]", DOLstring[loop], e_dol);  /* NJW's 1st E band Intensity map */
    else if (useA == 0) strcat( DOLstring_1[loop], "[1]");                       /* CBJ */
    else if ((useA == 1) || (numHDU <= 3)) strcat( DOLstring_1[loop], "[1]");    /* NL's raw images */
    else if ((useA == 2) && (numHDU > 3)){                                       /* NL's residual skymaps */
      strcat( DOLstring_1[loop], "[3]");
/*      ext[loop] = 2;                    */
    }
    else if ((useA == 3) && (numHDU > 4)){                                       /* NL's residual skymaps with sources */
      strcat( DOLstring_1[loop], "[4]");
      ext[loop] = 3;
    }

next_e_dol:    printf("The %d. DOL is : %s\n", loop+1, DOLstring_1[loop]);       /* NJW's FITS files may have more than 1 E-band */
    if (e_dol >= numHDU) printf("The energy interval is outside the selected energy range... -> excluded!\n");
    status = DALobjectOpen(DOLstring_1[loop], &dolPtr, status);
    if (status != ISDC_OK){
      printf("Cannot open DOL %s for image number %d\n", DOLstring_1[loop], loop+1);
      return status;
    }
    status1=ISDC_OK;
    status1 = DALattributeGetReal(dolPtr, "ONTIME", &exp_t[loop], NULL, NULL, status1);
    if(status1 != ISDC_OK) status = DALattributeGetReal(dolPtr, "EXPOSURE", &exp_t[loop], NULL, NULL, status);
      if(status != ISDC_OK){
      printf("Cannot read attribute ONTIME nor EXPOSURE for image number %d\n", loop+1);
      return( status );
      }
    if (exp_t[loop] <= 0.) printf("*** Warning: Image %d, EXPOSURE = %f ***\n", loop+1, exp_t[loop]);
    status = DALattributeGetReal(dolPtr, "CD1_1", &cd1_1[loop], NULL, NULL, status);
    if(status != ISDC_OK){
      printf("Cannot read attribute CD1_1 for image number %d\n", loop+1);
      return( status );
    }
    status = DALattributeGetReal(dolPtr, "CD2_1", &cd2_1[loop], NULL, NULL, status);
    if(status != ISDC_OK){
      printf("Cannot read attribute CD2_1 for image number %d\n", loop+1);
      return( status );
    }
    status = DALattributeGetReal(dolPtr, "CD1_2", &cd1_2[loop], NULL, NULL, status);
    if(status != ISDC_OK){
      printf("Cannot read attribute CD1_2 for image number %d\n", loop+1);
      return( status );
    }
    status = DALattributeGetReal(dolPtr, "CD2_2", &cd2_2[loop], NULL, NULL, status);
    if(status != ISDC_OK){
      printf("Cannot read attribute CD2_2 for image number %d\n", loop+1);
      return( status );
    }
    status = DALattributeGetReal(dolPtr, "CRVAL1", &crval1[loop], NULL, NULL, status);
    if(status != ISDC_OK){
      printf("Cannot read attribute CRVAL1 for image number %d\n", loop+1);
      return( status );
    }
    status = DALattributeGetReal(dolPtr, "CRVAL2", &crval2[loop], NULL, NULL, status);
    if(status != ISDC_OK){
      printf("Cannot read attribute CRVAL2 for image number %d\n", loop+1);
      return( status );
    }
    status = DALattributeGetReal(dolPtr, "CRPIX1", &crpix1[loop], NULL, NULL, status);
    if(status != ISDC_OK){
      printf("Cannot read attribute CRPIX1 for image number %d\n", loop+1);
      return( status );
    }
    status = DALattributeGetReal(dolPtr, "CRPIX2", &crpix2[loop], NULL, NULL, status);
    if(status != ISDC_OK){
      printf("Cannot read attribute CRPIX2 for image number %d\n", loop+1);
      return( status );
    }
    status = DALattributeGetReal(dolPtr, "E_MIN", &e_min[loop], NULL, NULL, status);
    if(status != ISDC_OK){
      printf("Cannot read attribute E_MIN for image number %d\n", loop+1);
      return( status );
    }
    status = DALattributeGetReal(dolPtr, "E_MAX", &e_max[loop], NULL, NULL, status);
    if(status != ISDC_OK){
      printf("Cannot read attribute E_MAX for image number %d\n", loop+1);
      return( status );
    }
    if (useE){ status = DALattributeGetInt(dolPtr, "ECODE", &ecode[loop], NULL, NULL, status);
    if(status != ISDC_OK){
      printf("Cannot read attribute ECODE for image number %d\n", loop+1);
      return( status );
    } }
    status = DALobjectClose(dolPtr, DAL_SAVE, status);
    if (status != ISDC_OK){
      printf("Cannot close DOL for image number %d\n", loop+1);
      return status;
    }

    if (useE){
      if ( (ecode[loop]<ecode_min) || (ecode[loop]>ecode_max) ){
        printf("The energy interval is outside the selected energy range... -> excluded!\n");
        cnt++;
        continue;
      }
    }
    if ( (e_min[loop]<emin) || (e_max[loop]>emax) ){
      if (useA < 0){
        e_dol += 2;
        sprintf(DOLstring_1[loop], "%s[%d]", DOLstring[loop], e_dol);
        goto next_e_dol;
      } 
      else {
        printf("The energy interval is outside the selected energy range... -> excluded!\n");
        cnt++;
        continue;
      }
    }
    if (e_min[loop] < min_e) min_e = e_min[loop];
    if (e_max[loop] > max_e) max_e = e_max[loop];
  }
  numImages = loop+1;

  fclose (fp1);

/* Define the output mosaic skymap */
  crval1_min = 360.;
  crval1_max = 0.;
  crval2_min = 90.;
  crval2_max = -90.;
  for(loop=0; loop<numImages; loop++){
    if (crval1[loop] < crval1_min) crval1_min = crval1[loop];
    if (crval2[loop] < crval2_min) crval2_min = crval2[loop];
    if (crval1[loop] > crval1_max) crval1_max = crval1[loop];
    if (crval2[loop] > crval2_max) crval2_max = crval2[loop];
  }
  if ((crval1_max-crval1_min) >= 180.) crval1_max -= 360.;
  if (crval1_max < crval1_min){
    crval1_min = crval1_max;  
    for(loop=0; loop<numImages; loop++)
      if ((crval1[loop] > crval1_max) && crval1[loop] <= 180.) crval1_max = crval1[loop];
  }
  if (diamout > 0.){
    d_a = crval1_max-crval1_min;
    d_b = 90.-crval2_min;
    d_c = 90.-crval2_max;
    d_D = cos(d_b*dr)*cos(d_c*dr)+sin(d_b*dr)*sin(d_c*dr)*cos(d_a*dr);
    if ( (acos(d_D)/dr) >= diamout ){
      printf("*** Warning: total separation angle larger than %f degrees, some images may not be included ***\n", diamout);
      printf("Largest separation angle is: %f degrees\n", acos(d_D)/dr);
    }
  }
  if (centerA < 0.){
    crval1_o = 0.5 * (crval1_min + crval1_max);
    if (crval1_o < 0.) crval1_o+= 360.;
    crval2_o = 0.5 * (crval2_min + crval2_max);
  }
  else {
    crval1_o = centerA;
    crval2_o = centerD;
  }
  printf("Center of output mosaic skyimage: RA=%f, DEC=%f\n", crval1_o, crval2_o);

/* Define transformation from sky coord. to FITS */
  /*  cdelt = 0.03 ;  degrees per pixel*/
  if ((diamout > 0.) && (cdelt > 0.)){
    dimx = (int)(diamout / cdelt);
    dimy = (int)(diamout / cdelt);
  }
  crpix1_o = (float)(dimx)/2.;
  crpix2_o = (float)(dimy)/2.;
  cd1_1_o = - cdelt;
  cd1_2_o = 0.0;
  cd2_1_o = 0.0;
  cd2_2_o = cdelt;
  ra_o = crval1_o;
  de_o = crval2_o;
  t1_o[0][0] = cos(ra_o*dr);
  t1_o[1][0] = sin(ra_o*dr);
  t1_o[2][0] = 0.;
  t1_o[0][1] = -sin(ra_o*dr);
  t1_o[1][1] = cos(ra_o*dr);
  t1_o[2][1] = 0.;
  t1_o[0][2] = 0.;
  t1_o[1][2] = 0.;
  t1_o[2][2] = 1.;
  t2_o[0][0] = cos(de_o*dr);
  t2_o[1][0] = 0.;
  t2_o[2][0] = sin(de_o*dr);
  t2_o[0][1] = 0.;
  t2_o[1][1] = 1.;
  t2_o[2][1] = 0.;
  t2_o[0][2] = -sin(de_o*dr);
  t2_o[1][2] = 0.;
  t2_o[2][2] = cos(de_o*dr);
  for (i=0; i<3; i++)
    for (j=0; j<3; j++){
      t_o[i][j] = 0.;
      for (k=0; k<3; k++) t_o[i][j] += t1_o[i][k]*t2_o[k][j];
    }
  for (i=0; i<3; i++)
    for (j=0; j<3; j++) o_t[i][j] = t_o[j][i];

/* COMPOSE MOSAIC BY LOOPING OVER SKY IMAGES TO ADD */
  cnt = 0;
  for (loop=0; loop<numImages; loop++){
    if ( diamout > 0.){
      d_a = crval1[loop]-crval1_o;
      d_b = 90.-crval2[loop];
      d_c = 90.-crval2_o;
      d_D = cos(d_b*dr)*cos(d_c*dr)+sin(d_b*dr)*sin(d_c*dr)*cos(d_a*dr);
      if ( (acos(d_D)/dr-radius) >= diamout/2. ){
        printf(" Excluding %s\n", DOLstring[loop]);
        cnt++;
        continue;
      }
    }
    for (i=0; i<MOSAICSIZE; i++) mark[i] = 0;
  /* Define transformation from FITS to sky coord. */
    ra_x = crval1[loop];
    de_x = crval2[loop];
    t1[0][0] = cos(ra_x*dr);
    t1[1][0] = sin(ra_x*dr);
    t1[2][0] = 0.;
    t1[0][1] = -sin(ra_x*dr);
    t1[1][1] = cos(ra_x*dr);
    t1[2][1] = 0.;
    t1[0][2] = 0.;
    t1[1][2] = 0.;
    t1[2][2] = 1.;
    t2[0][0] = cos(de_x*dr);
    t2[1][0] = 0.;
    t2[2][0] = sin(de_x*dr);
    t2[0][1] = 0.;
    t2[1][1] = 1.;
    t2[2][1] = 0.;
    t2[0][2] = -sin(de_x*dr);
    t2[1][2] = 0.;
    t2[2][2] = cos(de_x*dr);
    for (i=0; i<3; i++)
      for (j=0; j<3; j++){
        t[i][j] = 0.;
        for (k=0; k<3; k++) t[i][j] += t1[i][k]*t2[k][j];
      }

  /* Read skymaps inputs */
    status = DALobjectOpen(DOLstring_1[loop], &dolPtr, status);
    if (status != ISDC_OK){
      printf("Cannot open DOL %s for image number %d\n", DOLstring_1[loop], loop+1);
      return status;
    }
    arrayPtr = pvalue;
    status = DALarrayGetSection(dolPtr, numAxes, startValues, endValues, &DALtype, &numValues, arrayPtr, status);
    if (status != ISDC_OK){
      printf("Cannot get section pvalues for image number %d\n", loop+1);
      return status;
    }
    status = DALobjectClose(dolPtr, DAL_SAVE, status);
    if (status != ISDC_OK){
      printf("Cannot close DOL for image number %d\n", loop+1);
      return status;
    }

    if (ext[loop] == 2){  /* Use of NL's filtering function for large scale artefacts in images without strong sources */
      for (p=0; p<(int)(numValues); p++) value_in[p] = (float)(pvalue[p]);
      arrayPtr = value_out;
      clean(value_in, arrayPtr,1);
      for (p=0; p<(int)(numValues); p++) pvalue[p] = (double)(value_out[p]);
    }

    /* Get variance maps */
    strcpy(DOLstring_2, DOLstring[loop]);
    if (useA > 0) strcat(DOLstring_2, "[2]");                                    /* Extension is [2] for NL's variance maps */
    else if (useA == 0) strcat(DOLstring_2, "[2]");                              /* idem for CBJ's */
    else if (useA < 0) sprintf(DOLstring_2, "%s[%d]", DOLstring[loop], e_dol+1); /* NJW's variance map next to skymap for each E-band */
    printf("Variance map is in : %s\n", DOLstring_2);
    status = DALobjectOpen(DOLstring_2, &dolPtr, status);
    if (status != ISDC_OK){
      printf("Cannot open DOL %s for image number %d\n", DOLstring_2, loop+1);
      return status;
    }
    arrayPtr = variance;
    status = DALarrayGetSection(dolPtr, numAxes, startValues, endValues, &DALtype, &numValues, arrayPtr, status);
    if (status != ISDC_OK){
      printf("Cannot get section variance for image number %d\n", loop);
      return status;
    }
    status = DALobjectClose(dolPtr, DAL_SAVE, status);
    if (status != ISDC_OK){
      printf("Cannot close DOL for image number %d\n", loop+1);
      return status;
    }
    if (useA > 0){  /* Use of NL's filtering function for large scale artefacts in variance maps */
      for (p=0; p<(int)(numValues); p++) value_in[p] = (float)(variance[p]);
      arrayPtr = value_out;
      clean(value_in, arrayPtr,0);
      for (p=0; p<(int)(numValues); p++) variance[p] = (double)(value_out[p]);
    }
     
    if (useA == 0){       /* Get CBJ's special vignetting correcting factors */
      if (xdim == 262){
        strcpy(vign_s, vign_d);
        strcat(vign_s, "cbj_vig.dat");
        if( (vignet = fopen(vign_s,"rt")) == NULL ) {
          printf("Could not open the file named: %s\n", vign_s);
          return(0);
        }
        j = 0;
        while (j<xdim*ydim){
          for (i=0; i<43; i++){
            fgets(vign_s, 99, vignet);
            sscanf(vign_s, "%d %d %d %d %d %d", &pvalcor[j++],&pvalcor[j++],&pvalcor[j++],&pvalcor[j++],&pvalcor[j++],&pvalcor[j++]);
          }
          fgets(vign_s, 99, vignet);
          sscanf(vign_s, "%d %d %d %d", &pvalcor[j++],&pvalcor[j++],&pvalcor[j++],&pvalcor[j++]);
        }
        fclose(vignet);
      }
      else if (xdim == 525){
        strcpy(vign_s, vign_d);
        strcat(vign_s, "bckim_1_5.txt");
        if( (vignet = fopen(vign_s,"rt")) == NULL ) {
          printf("Could not open the file named: %s\n", vign_s);
          return(0);
        }
        j = 0;
        while (j<xdim*ydim){
          for (i=0; i<131; i++){
            fgets(vign_s, 99, vignet);
            sscanf(vign_s, "%lf %lf %f %lf", &pvalcor_f[j++],&pvalcor_f[j++],&pvalcor_f[j++],&pvalcor_f[j++]);
          }
          fgets(vign_s, 99, vignet);
          sscanf(vign_s, "%lf", &pvalcor_f[j++]);
        }
        fclose(vignet);
      }
    }

    for (p=0; p<(int)(numValues); p++){
      pvalue[p] = pvalue[p] * reduc;  /* Divide the pixel values by the nb of pixel subdivisions */
      variance[p] = variance[p] * reduc; /* Divide the variances by the nb of pixel subdivisions */
      if ((useA == 0) && (xdim == 262)) pvalue[p] = pvalue[p] * pvalcor[p];
      else if ((useA == 0) && (xdim == 525)) pvalue[p] = pvalue[p] * pvalcor_f[p];
    }

    /* Loop over individual pixels */
    for (p=0; p<xdim; p++){
      for (q=0; q<ydim; q++){
        index1 = q*xdim+p;
        pval_i[p][q] = isnan(pvalue[index1]) ? 0.0 : pvalue[index1];
        var_i[p][q] = variance[index1] +100.;  /* Add a threshold to the variance to avoid too high weights */
        exp_i[p][q] = exp_t[loop]*reduc * vign[p][q]; /* Compute effective exposure with vignetting-function normalised to 1 */
        yz[0] = cd1_1[loop]*(p+1-crpix1[loop]) + cd1_2[loop]*(q+1-crpix2[loop]);
        yz[1] = cd2_1[loop]*(p+1-crpix1[loop]) + cd2_2[loop]*(q+1-crpix2[loop]);
        if (sqrt(yz[0]*yz[0]+yz[1]*yz[1]) > radius) continue;
        /* Loop over pixels subdivisions  */
        for (psub=-isub; psub<=isub; psub++){
        for (qsub=-isub; qsub<=isub; qsub++){
          dp = psub*dsub;
          dq = qsub*dsub;
          yz[0] = cd1_1[loop]*(p+dp+1-crpix1[loop]) + cd1_2[loop]*(q+dq+1-crpix2[loop]);
          yz[1] = cd2_1[loop]*(p+dp+1-crpix1[loop]) + cd2_2[loop]*(q+dq+1-crpix2[loop]);
          for(k=0; k<2; k++) ayz[k] = atan(yz[k]*dr);
          for(k=1; k<3; k++) vi[k] = sin(ayz[k-1]);
          vi[0] = sqrt(1. -vi[1]*vi[1]-vi[2]*vi[2]);
          for (i=0; i<3; i++){
            ve[i] = 0.;
            for (j=0; j<3; j++) ve[i] += t[i][j]*vi[j];
          }
/*
          dec = asin(ve[2])/dr;
          ra = atan2(ve[1],ve[0])/dr + 360.;
          if (ra > 360.) ra -= 360.;
*/
          /* Where does this position belong in mosaic image? */
          for (i=0; i<3; i++){
            vi_o[i] = 0.;
            for (j=0; j<3; j++) vi_o[i] += o_t[i][j]*ve[j];
          }
          for (k=0; k<2; k++){
            ayz[k] = asin(vi_o[k+1]);
            yz[k] = tan(ayz[k]) / dr;
          }
          pix[0] = (int)((-1./cdelt)*yz[0]+ crpix1_o -1.);
          pix[1] = (int)((1./cdelt)*yz[1] + crpix2_o -1.);
          index2 = pix[1]*dimx+pix[0];
          if ((pix[0]>=0) && (pix[0]<dimx) && (pix[1]>=0) && (pix[1]<dimy)){
/*      COMPUTE WEIGTHS AND UPDATE MOSAIC SKYMAP 
            if (exp_o[index2] > 0.)
             wei = (exp_i[p][q]/var_i[p][q]) / (exp_o[index2]/var_o[index2]);
            else wei = 1.0;
            if (wei <= 1.0){
              pval_o[index2] += wei * pval_i[p][q];
              exp_o[index2] += wei * exp_i[p][q];
              var_o[index2] += wei * wei * var_i[p][q];
            }
            else{
              wei = 1./wei; 
              pval_o[index2] = pval_i[p][q] + wei * pval_o[index2];
              if ((exp_i[p][q]+wei*exp_o[index2]) > exp_o[index2]) exp_o[index2] = exp_i[p][q] + wei * exp_o[index2];
              var_o[index2] = var_i[p][q] + wei * wei * var_o[index2];
            }
            weight[index2] += wei;
            weigh2[index2] += wei*wei;
*/          wei = (exp_i[p][q]/var_i[p][q]);
            pval_o[index2] += wei * pval_i[p][q];
            exp_o[index2] += wei * exp_i[p][q];
            var_o[index2] += wei * wei * var_i[p][q];
            weight[index2] += wei;
            weigh2[index2] += wei*wei;
            exp_s[index2] += exp_i[p][q];
            var_s[index2] += var_i[p][q];
            rat_s[index2] = var_s[index2]/exp_s[index2];
            if (mark[index2]==0){
              t_exp[index2] += exp_t[loop]; /* Compute raw exposure times */
              v_exp[index2] += exp_t[loop] * vign[p][q]; /* Compute effective exposure times */
              mark[index2] ++;
              if ((index2 == dimx*(dimy/2-1)+dimx/2-1) && (centerA >=0.)) rwei[loop] = wei;
            }
          }
        }
        }
      }
    }
  } /* End loop over total number of sky images to add */
  for (index2=0; index2<MOSAICSIZE; index2++){
    pval_o[index2] = pval_o[index2]*rat_s[index2];
    exp_o[index2] = exp_o[index2]*rat_s[index2];
    var_o[index2] = var_o[index2]*rat_s[index2]*rat_s[index2];
    weight[index2] = weight[index2]*rat_s[index2];
    weigh2[index2] = weigh2[index2]*rat_s[index2]*rat_s[index2];
    if (!pval_o[index2]) pval_o[index2] = NaN;
    if (!exp_o[index2]) exp_o[index2] = NaN;
    if (!var_o[index2]) var_o[index2] = NaN;
    if (!t_exp[index2]) t_exp[index2] = NaN;
    if (!v_exp[index2]) v_exp[index2] = NaN;
    if (!weight[index2]) weight[index2] = NaN;
    if (!weigh2[index2]) weigh2[index2] = NaN;
  }

  numImages -= cnt;

  if (centerA >=0.){
    printf("-> List of relative weights in ranked_weights.txt\n");
    fp1 = fopen("ranked_weights.txt", "wt");
    fprintf(fp1, "********************************************************************************\n");
    fprintf(fp1, "		       Ordered list of the input images \n");
    fprintf(fp1, "		    as a function of their relative weights \n");
    fprintf(fp1, "	      at the given position: RA=%f  DEC=%f \n", centerA, centerD);
    fprintf(fp1, "********************************************************************************\n\n");
    p = 0;
    for (overloop=0; overloop<numImages; overloop++){
      if (rwei[overloop] < 0.){
        fprintf(fp1, "The given position is outside the selected input radius in %s!\n", DOLstring[overloop]);
        p++;
      }
      for (loop=0; loop<numImages; loop++){
        for(i=0; i<numImages; i++) if (loopM[i] == loop) testM=1;
          if (testM){
            if (!rank[loop]) rank[loop] = overloop;
            testM = 0;
            continue;
          }
        if (rwei[loop]>=rweiMax){
          rweiMax = rwei[loop];
          k = loop;
        }
      }
      rweiMax = 0.;
      loopM[overloop] = k;
    }
    fprintf(fp1, "\n");
    for (loop=0; loop<numImages-p; loop++){
      if (!rank[loop]) rank[loop] = numImages-p;
      fprintf(fp1, "The input image %s\n", DOLstring_1[loopM[loop]]);
      fprintf(fp1, "has rank %d with a relative weight equal to: %f\n\n", rank[loopM[loop]], rwei[loopM[loop]]);
    }
    fclose (fp1);
  }

/* Generate output Mosaic fits file */
  endValues[0] = dimx;
  endValues[1] = dimy;
  if (plot_T){
    sprintf(fitstring, "OBS_TIME_%d", numImages);
    status = DALobjectCreate(fitstring, DAL_DISK, template, &timPtr, status);
    if (status != ISDC_OK){
      printf("Cannot create OBS_TIME object, status = %d\n", status);
      return status;
    }

    printf("-> Plot exposure Time map in %s.fits[2]\n", fitstring);
    status = DAL3GENindexCreateMember(timPtr, NULL, NULL, &texPtr, status);
    if (status != ISDC_OK){
      printf("DAL3GEN: Cannot create member for t_exp object\n");
      return status;
    }
    status = DALarrayModStruct(texPtr, DALtype, numAxes, endValues, status);
    if( status != ISDC_OK )  printf("DALarrayModStruct status : %d\n", status);
    status += DALattributePutChar(texPtr, "CTYPE1", "RA---TAN","1962","Projection name", status);
    status += DALattributePutChar(texPtr, "CTYPE2", "DEC--TAN","1962 ","Projection name", status);
    status += DALattributePutReal(texPtr, "CRVAL1", crval1_o,"deg","Sky coord. ref. 1. axis", status);
    status += DALattributePutReal(texPtr, "CRVAL2", crval2_o,"deg","Sky coord. ref. 2. axis", status);
    status += DALattributePutReal(texPtr, "CD1_1", cd1_1_o,"deg", "Transformation", status);
    status += DALattributePutReal(texPtr, "CD2_2", cd2_2_o,"deg", "Transformation", status);
    status += DALattributePutReal(texPtr, "CD1_2", cd1_2_o,"deg", "Transformation", status);
    status += DALattributePutReal(texPtr, "CD2_1", cd2_1_o,"deg", "Transformation", status);
    status += DALattributePutReal(texPtr, "CRPIX1", crpix1_o,"pixel", "Sky pixel ref. 1. axis", status);
    status += DALattributePutReal(texPtr, "CRPIX2", crpix2_o,"pixel", "Sky pixel ref. 2. axis", status);
    status += DALattributePutChar(texPtr, "IMATYPE", "OBS_TIME", "seconds", "The image shows observation times", status);
    arrayBuff = t_exp;
    status = DALarrayPutSection(texPtr, numAxes, startValues, endValues, DALtype, arrayBuff, status);
    if (status != ISDC_OK){
      printf("Cannot put section t_exp object\n");
      return status;
    }

    printf("-> Plot effective exposure Time map in %s.fits[3]\n", fitstring);
    status = DAL3GENindexCreateMember(timPtr, NULL, NULL, &extPtr, status);
    if (status != ISDC_OK){
      printf("DAL3GEN: Cannot create member for v_exp object\n");
      return status;
    }
    status = DALarrayModStruct(extPtr, DALtype, numAxes, endValues, status);
    if( status != ISDC_OK )  printf("DALarrayModStruct status : %d\n", status);
    status += DALattributePutChar(extPtr, "CTYPE1", "RA---TAN","1962","Projection name", status);
    status += DALattributePutChar(extPtr, "CTYPE2", "DEC--TAN","1962 ","Projection name", status);
    status += DALattributePutReal(extPtr, "CRVAL1", crval1_o,"deg","Sky coord. ref. 1. axis", status);
    status += DALattributePutReal(extPtr, "CRVAL2", crval2_o,"deg","Sky coord. ref. 2. axis", status);
    status += DALattributePutReal(extPtr, "CD1_1", cd1_1_o,"deg", "Transformation", status);
    status += DALattributePutReal(extPtr, "CD2_2", cd2_2_o,"deg", "Transformation", status);
    status += DALattributePutReal(extPtr, "CD1_2", cd1_2_o,"deg", "Transformation", status);
    status += DALattributePutReal(extPtr, "CD2_1", cd2_1_o,"deg", "Transformation", status);
    status += DALattributePutReal(extPtr, "CRPIX1", crpix1_o,"pixel", "Sky pixel ref. 1. axis", status);
    status += DALattributePutReal(extPtr, "CRPIX2", crpix2_o,"pixel", "Sky pixel ref. 2. axis", status);
    status += DALattributePutChar(extPtr, "IMATYPE", "EXP_TIME", "seconds", "The image shows exposure times", status);
    arrayBuff = v_exp;
    status = DALarrayPutSection(extPtr, numAxes, startValues, endValues, DALtype, arrayBuff, status);
    if (status != ISDC_OK){
      printf("Cannot put section v_exp object\n");
      return status;
    }

    status = DALobjectClose(timPtr, DAL_SAVE, status);
    if (status != ISDC_OK){
      printf("Cannot close object!\n");
      return status;
    }
  }

  if (!suffix) sprintf(fitstring, "%s", mosaic);
  else sprintf(fitstring, "%s_%d_%d", mosaic, useA, numImages);
  status = DALobjectCreate(fitstring, DAL_DISK, template, &skyPtr, status);
  if (status != ISDC_OK){
    printf("Cannot create MOSAIC object, status = %d\n", status);
    return status;
  }

  if (plot_I){
    nex++;
    printf("-> Plot raw Intensity map in %s.fits[%d]\n", fitstring, nex);
    status = DAL3GENindexCreateMember(skyPtr, NULL, NULL, &pvlPtr, status);
    if (status != ISDC_OK){
      printf("DAL3GEN: Cannot create member for pval_o object\n");
      return status;
    }
    status = DALarrayModStruct( pvlPtr, DALtype, numAxes, endValues, status);
    if( status != ISDC_OK )  printf("DALarrayModStruct status : %d\n", status);
    status += DALattributePutChar( pvlPtr, "CTYPE1", "RA---TAN","1962","Projection name", status);
    status += DALattributePutChar( pvlPtr, "CTYPE2", "DEC--TAN","1962 ","Projection name", status);
    status += DALattributePutReal( pvlPtr, "CRVAL1", crval1_o,"deg","Sky coord. ref. 1. axis", status);
    status += DALattributePutReal( pvlPtr, "CRVAL2", crval2_o,"deg","Sky coord. ref. 2. axis", status);
    status += DALattributePutReal( pvlPtr, "CD1_1", cd1_1_o,"deg", "Transformation", status);
    status += DALattributePutReal( pvlPtr, "CD2_2", cd2_2_o,"deg", "Transformation", status);
    status += DALattributePutReal( pvlPtr, "CD1_2", cd1_2_o,"deg", "Transformation", status);
    status += DALattributePutReal( pvlPtr, "CD2_1", cd2_1_o,"deg", "Transformation", status);
    status += DALattributePutReal( pvlPtr, "CRPIX1", crpix1_o,"pixel", "Sky pixel ref. 1. axis", status);
    status += DALattributePutReal( pvlPtr, "CRPIX2", crpix2_o,"pixel", "Sky pixel ref. 2. axis", status);
    status += DALattributePutReal( pvlPtr, "E_MIN", min_e,"keV", "Lower Energy limit", status);
    status += DALattributePutReal( pvlPtr, "E_MAX", max_e,"keV", "Upper Energy limit", status);
    status += DALattributePutChar( pvlPtr, "IMATYPE", "INTENSITY", " ", "The image shows intensity", status);
    arrayBuff = pval_o;
    status = DALarrayPutSection(pvlPtr, numAxes, startValues, endValues, DALtype, arrayBuff, status);
    if (status != ISDC_OK){
      printf("Cannot put section pval_o object\n");
      return status;
    }
  }
  
  if (plot_E){
    nex++;
    printf("-> Plot effective Exposure map in %s.fits[%d]\n", fitstring, nex);
    status = DAL3GENindexCreateMember(skyPtr, NULL, NULL, &expPtr, status);
    if (status != ISDC_OK){
      printf("DAL3GEN: Cannot create member for exp_o object\n");
      return status;
    }
    status = DALarrayModStruct( expPtr, DALtype, numAxes, endValues, status);
    if( status != ISDC_OK )  printf("DALarrayModStruct status : %d\n", status);
    status += DALattributePutChar( expPtr, "CTYPE1", "RA---TAN","1962","Projection name", status);
    status += DALattributePutChar( expPtr, "CTYPE2", "DEC--TAN","1962 ","Projection name", status);
    status += DALattributePutReal( expPtr, "CRVAL1", crval1_o,"deg","Sky coord. ref. 1. axis", status);
    status += DALattributePutReal( expPtr, "CRVAL2", crval2_o,"deg","Sky coord. ref. 2. axis", status);
    status += DALattributePutReal( expPtr, "CD1_1", cd1_1_o,"deg", "Transformation", status);
    status += DALattributePutReal( expPtr, "CD2_2", cd2_2_o,"deg", "Transformation", status);
    status += DALattributePutReal( expPtr, "CD1_2", cd1_2_o,"deg", "Transformation", status);
    status += DALattributePutReal( expPtr, "CD2_1", cd2_1_o,"deg", "Transformation", status);
    status += DALattributePutReal( expPtr, "CRPIX1", crpix1_o,"pixel", "Sky pixel ref. 1. axis", status);
    status += DALattributePutReal( expPtr, "CRPIX2", crpix2_o,"pixel", "Sky pixel ref. 2. axis", status);
    status += DALattributePutReal( expPtr, "E_MIN", min_e,"keV", "Lower Energy limit", status);
    status += DALattributePutReal( expPtr, "E_MAX", max_e,"keV", "Upper Energy limit", status);
    status += DALattributePutChar( expPtr, "IMATYPE", "EXPOSURE", " ", "The image shows effective exposure", status);
    arrayBuff = exp_o;
    status = DALarrayPutSection(expPtr, numAxes, startValues, endValues, DALtype, arrayBuff, status);
    if (status != ISDC_OK){
      printf("Cannot put section exp_o object\n");
      return status;
    }
  }

  if (plot_V){
    nex++;
    printf("-> Plot Variance map in %s.fits[%d]\n", fitstring, nex);
    status = DAL3GENindexCreateMember(skyPtr, NULL, NULL, &varPtr, status);
    if (status != ISDC_OK){
      printf("DAL3GEN: Cannot create member for var_o object\n");
      return status;
    }
    status = DALarrayModStruct( varPtr, DALtype, numAxes, endValues, status);
    if( status != ISDC_OK )  printf("DALarrayModStruct status : %d\n", status);
    status += DALattributePutChar( varPtr, "CTYPE1", "RA---TAN","1962","Projection name", status);
    status += DALattributePutChar( varPtr, "CTYPE2", "DEC--TAN","1962 ","Projection name", status);
    status += DALattributePutReal( varPtr, "CRVAL1", crval1_o,"deg","Sky coord. ref. 1. axis", status);
    status += DALattributePutReal( varPtr, "CRVAL2", crval2_o,"deg","Sky coord. ref. 2. axis", status);
    status += DALattributePutReal( varPtr, "CD1_1", cd1_1_o,"deg", "Transformation", status);
    status += DALattributePutReal( varPtr, "CD2_2", cd2_2_o,"deg", "Transformation", status);
    status += DALattributePutReal( varPtr, "CD1_2", cd1_2_o,"deg", "Transformation", status);
    status += DALattributePutReal( varPtr, "CD2_1", cd2_1_o,"deg", "Transformation", status);
    status += DALattributePutReal( varPtr, "CRPIX1", crpix1_o,"pixel", "Sky pixel ref. 1. axis", status);
    status += DALattributePutReal( varPtr, "CRPIX2", crpix2_o,"pixel", "Sky pixel ref. 2. axis", status);
    status += DALattributePutReal( varPtr, "E_MIN", min_e,"keV", "Lower Energy limit", status);
    status += DALattributePutReal( varPtr, "E_MAX", max_e,"keV", "Upper Energy limit", status);
    status += DALattributePutChar( varPtr, "IMATYPE", "VARIANCE", " ", "The image shows variance", status);
    arrayBuff = var_o;
    status = DALarrayPutSection(varPtr, numAxes, startValues, endValues, DALtype, arrayBuff, status);
    if (status != ISDC_OK){
      printf("Cannot put section var_o object\n");
      return status;
    }
  }

  if (plot_W){
    nex++;
    printf("-> Plot sum of Weights map in %s.fits[%d]\n", fitstring, nex);
    status = DAL3GENindexCreateMember(skyPtr, NULL, NULL, &sw_Ptr, status);
    if (status != ISDC_OK){
      printf("DAL3GEN: Cannot create member for weight object\n");
      return status;
    }
    status = DALarrayModStruct( sw_Ptr, DALtype, numAxes, endValues, status);
    if( status != ISDC_OK )  printf("DALarrayModStruct status : %d\n", status);
    status += DALattributePutChar( sw_Ptr, "CTYPE1", "RA---TAN","1962","Projection name", status);
    status += DALattributePutChar( sw_Ptr, "CTYPE2", "DEC--TAN","1962 ","Projection name", status);
    status += DALattributePutReal( sw_Ptr, "CRVAL1", crval1_o,"deg","Sky coord. ref. 1. axis", status);
    status += DALattributePutReal( sw_Ptr, "CRVAL2", crval2_o,"deg","Sky coord. ref. 2. axis", status);
    status += DALattributePutReal( sw_Ptr, "CD1_1", cd1_1_o,"deg", "Transformation", status);
    status += DALattributePutReal( sw_Ptr, "CD2_2", cd2_2_o,"deg", "Transformation", status);
    status += DALattributePutReal( sw_Ptr, "CD1_2", cd1_2_o,"deg", "Transformation", status);
    status += DALattributePutReal( sw_Ptr, "CD2_1", cd2_1_o,"deg", "Transformation", status);
    status += DALattributePutReal( sw_Ptr, "CRPIX1", crpix1_o,"pixel", "Sky pixel ref. 1. axis", status);
    status += DALattributePutReal( sw_Ptr, "CRPIX2", crpix2_o,"pixel", "Sky pixel ref. 2. axis", status);
    status += DALattributePutReal( sw_Ptr, "E_MIN", min_e,"keV", "Lower Energy limit", status);
    status += DALattributePutReal( sw_Ptr, "E_MAX", max_e,"keV", "Upper Energy limit", status);
    status += DALattributePutChar( sw_Ptr, "IMATYPE", "WEIGHTS", " ", "The image shows sum of weights", status);
    arrayBuff = weight;
    status = DALarrayPutSection(sw_Ptr, numAxes, startValues, endValues, DALtype, arrayBuff, status);
    if (status != ISDC_OK){
      printf("Cannot put section weight object\n");
      return status;
    }
  }

  if (plot_W2){
    nex++;
    printf("-> Plot sum of weights^2 map in %s.fits[%d]\n", fitstring, nex);
    status = DAL3GENindexCreateMember(skyPtr, NULL, NULL, &sw2Ptr, status);
    if (status != ISDC_OK){
      printf("DAL3GEN: Cannot create member for weigh2 object\n");
      return status;
    }
    status = DALarrayModStruct( sw2Ptr, DALtype, numAxes, endValues, status);
    if( status != ISDC_OK )  printf("DALarrayModStruct status : %d\n", status);
    status += DALattributePutChar( sw2Ptr, "CTYPE1", "RA---TAN","1962","Projection name", status);
    status += DALattributePutChar( sw2Ptr, "CTYPE2", "DEC--TAN","1962 ","Projection name", status);
    status += DALattributePutReal( sw2Ptr, "CRVAL1", crval1_o,"deg","Sky coord. ref. 1. axis", status);
    status += DALattributePutReal( sw2Ptr, "CRVAL2", crval2_o,"deg","Sky coord. ref. 2. axis", status);
    status += DALattributePutReal( sw2Ptr, "CD1_1", cd1_1_o,"deg", "Transformation", status);
    status += DALattributePutReal( sw2Ptr, "CD2_2", cd2_2_o,"deg", "Transformation", status);
    status += DALattributePutReal( sw2Ptr, "CD1_2", cd1_2_o,"deg", "Transformation", status);
    status += DALattributePutReal( sw2Ptr, "CD2_1", cd2_1_o,"deg", "Transformation", status);
    status += DALattributePutReal( sw2Ptr, "CRPIX1", crpix1_o,"pixel", "Sky pixel ref. 1. axis", status);
    status += DALattributePutReal( sw2Ptr, "CRPIX2", crpix2_o,"pixel", "Sky pixel ref. 2. axis", status);
    status += DALattributePutReal( sw2Ptr, "E_MIN", min_e,"keV", "Lower Energy limit", status);
    status += DALattributePutReal( sw2Ptr, "E_MAX", max_e,"keV", "Upper Energy limit", status);
    status += DALattributePutChar( sw2Ptr, "IMATYPE", "WEIGHT2", " ", "The image shows sum of weights^2", status);
    arrayBuff = weigh2;
    status = DALarrayPutSection(sw2Ptr, numAxes, startValues, endValues, DALtype, arrayBuff, status);
    if (status != ISDC_OK){
      printf("Cannot put section weigh2 object\n");
      return status;
    }
  }

  for (index2=0; index2<MOSAICSIZE; index2++)
    if (exp_o[index2] != 0.){
      pval_o[index2] = pval_o[index2]/exp_o[index2];              /* Actual quantities to plot */
      var_o[index2] = sqrt(var_o[index2])/exp_o[index2];          /* pv/exp and  sqrt(var)/exp */
    }

  if (plot_SoE){
    nex++;
    printf("-> Plot Signal/Exposure map in %s.fits[%d]\n", fitstring, nex);
    status = DAL3GENindexCreateMember(skyPtr, NULL, NULL, &mapPtr, status);
    if (status != ISDC_OK){
      printf("DAL3GEN: Cannot create member for new pval_o object\n");
      return status;
    }
    status = DALarrayModStruct( mapPtr, DALtype, numAxes, endValues, status);
    if( status != ISDC_OK )  printf("DALarrayModStruct status : %d\n", status);
    status += DALattributePutChar( mapPtr, "CTYPE1", "RA---TAN","1962","Projection name", status);
    status += DALattributePutChar( mapPtr, "CTYPE2", "DEC--TAN","1962 ","Projection name", status);
    status += DALattributePutReal( mapPtr, "CRVAL1", crval1_o,"deg","Sky coord. ref. 1. axis", status);
    status += DALattributePutReal( mapPtr, "CRVAL2", crval2_o,"deg","Sky coord. ref. 2. axis", status);
    status += DALattributePutReal( mapPtr, "CD1_1", cd1_1_o,"deg", "Transformation", status);
    status += DALattributePutReal( mapPtr, "CD2_2", cd2_2_o,"deg", "Transformation", status);
    status += DALattributePutReal( mapPtr, "CD1_2", cd1_2_o,"deg", "Transformation", status);
    status += DALattributePutReal( mapPtr, "CD2_1", cd2_1_o,"deg", "Transformation", status);
    status += DALattributePutReal( mapPtr, "CRPIX1", crpix1_o,"pixel", "Sky pixel ref. 1. axis", status);
    status += DALattributePutReal( mapPtr, "CRPIX2", crpix2_o,"pixel", "Sky pixel ref. 2. axis", status);
    status += DALattributePutReal( mapPtr, "E_MIN", min_e,"keV", "Lower Energy limit", status);
    status += DALattributePutReal( mapPtr, "E_MAX", max_e,"keV", "Upper Energy limit", status);
    status += DALattributePutChar( mapPtr, "IMATYPE", "SIGNAL", " ", "The image shows signal/exposure", status);
    arrayBuff = pval_o;
    status = DALarrayPutSection(mapPtr, numAxes, startValues, endValues, DALtype, arrayBuff, status);
    if (status != ISDC_OK){
      printf("Cannot put section new pval_o object\n");
      return status;
    }
  }

  if (plot_RVE){
    nex++;
    printf("-> Plot SQRT(Variance)/Exposure map in %s.fits[%d]\n", fitstring, nex);
    status = DAL3GENindexCreateMember(skyPtr, NULL, NULL, &sqtPtr, status);
    if (status != ISDC_OK){
      printf("DAL3GEN: Cannot create member for new var_o object\n");
      return status;
    }
    status = DALarrayModStruct( sqtPtr, DALtype, numAxes, endValues, status);
    if( status != ISDC_OK )  printf("DALarrayModStruct status : %d\n", status);
    status += DALattributePutChar( sqtPtr, "CTYPE1", "RA---TAN","1962","Projection name", status);
    status += DALattributePutChar( sqtPtr, "CTYPE2", "DEC--TAN","1962 ","Projection name", status);
    status += DALattributePutReal( sqtPtr, "CRVAL1", crval1_o,"deg","Sky coord. ref. 1. axis", status);
    status += DALattributePutReal( sqtPtr, "CRVAL2", crval2_o,"deg","Sky coord. ref. 2. axis", status);
    status += DALattributePutReal( sqtPtr, "CD1_1", cd1_1_o,"deg", "Transformation", status);
    status += DALattributePutReal( sqtPtr, "CD2_2", cd2_2_o,"deg", "Transformation", status);
    status += DALattributePutReal( sqtPtr, "CD1_2", cd1_2_o,"deg", "Transformation", status);
    status += DALattributePutReal( sqtPtr, "CD2_1", cd2_1_o,"deg", "Transformation", status);
    status += DALattributePutReal( sqtPtr, "CRPIX1", crpix1_o,"pixel", "Sky pixel ref. 1. axis", status);
    status += DALattributePutReal( sqtPtr, "CRPIX2", crpix2_o,"pixel", "Sky pixel ref. 2. axis", status);
    status += DALattributePutReal( sqtPtr, "E_MIN", min_e,"keV", "Lower Energy limit", status);
    status += DALattributePutReal( sqtPtr, "E_MAX", max_e,"keV", "Upper Energy limit", status);
    status += DALattributePutChar( sqtPtr, "IMATYPE", "NOISE", " ", "The image shows sqrt(var)/exposure", status);
    arrayBuff = var_o;
    status = DALarrayPutSection(sqtPtr, numAxes, startValues, endValues, DALtype, arrayBuff, status);
    if (status != ISDC_OK){
      printf("Cannot put section new var_o object\n");
      return status;
    }
  }

  status = DALobjectClose(skyPtr, DAL_SAVE, status);
  if (status != ISDC_OK){
    printf("Cannot close object!\n");
    return status;
  }

  return 0;
}

/*--------------------------------------------------------------------------------------*/

void clean(float sky_r[skydim], float skyim[skydim], int bol)
/* this subroutine removes large scale structure from a skymap
   The input skymap is provided in the in the array sky_r
   The cleaned skymap is returned in the array skyim
   Niels Lund, September 2003
*/
{
   int i, j, n, k, max_j[3] = {-1, -1, -1}, min_j[2] = {100000, 100000};
   int indx[10] = {5, sky_ydim-4, sky_ydim+4, 2*sky_ydim-3, 2*sky_ydim+3,
                  3*sky_ydim-2, 3*sky_ydim+2, 4*sky_ydim-1, 4*sky_ydim+1, 5*sky_ydim};
   float b, v, max[2] = {-100000.0, -100000.0}, min[2] = {100000.0, 100000.0};
   
   for (i=5*sky_ydim; i<skydim-5*sky_ydim; i++) {
      for (j=0; j<10; j++) {
         v = sky_r[i+indx[j]];
         if (v > max[0]) {
            max[1] = max[0];
            max_j[1] = max_j[0];
            max[0] = v;
            max_j[0] = i+indx[j];
         } else {
            if (v > max[1]) {
               max[1] = v;
               max_j[1] = i+indx[j];
            }
         }
         if (v < min[0]) {
            min[1] = min[0];
            min_j[1] = min_j[0];
            min[0] = v;
            min_j[0] = i+indx[j];
         } else {
            if (v < min[1]) {
               min[1] = v;
               min_j[1] = i+indx[j];
            }
         }

         v = sky_r[i-indx[j]];
         if (v > max[0]) {
            max[1] = max[0];
            max_j[1] = max_j[0];
            max[0] = v;
            max_j[0] = i-indx[j];
         } else {
            if (v > max[1]) {
               max[1] = v;
               max_j[1] = i-indx[j];
            }
         }
         if (v < min[0]) {
            min[1] = min[0];
            min_j[1] = min_j[0];
            min[0] = v;
            min_j[0] = i+indx[j];
         } else {
            if (v < min[1]) {
               min[1] = v;
               min_j[1] = i+indx[j];
            }
         }
      }
   
      b = 0.0;
      n = 0;
      for (j=0; j<10; j++) {
         k = i+indx[j];
         if (k == max_j[0]) goto try2;
         if (k == max_j[1]) goto try2;
         if (k == min_j[0]) goto try2;
         if (k == min_j[1]) goto try2;
         b += sky_r[k];
         n++;
try2:    k = i-indx[j];
         if (k == max_j[0]) continue;
         if (k == max_j[1]) continue;
         if (k == min_j[0]) continue;
         if (k == min_j[1]) continue;
         b += sky_r[k];
         n++;
      }
      
      b /= (float)(n);
      
      if (bol == 1) skyim[i] = sky_r[i] - b;
      else if (bol == 0) skyim[i] = b;
      else printf("ERROR: bol = %d\n", bol);
   }
   
   return;
}

/* ********************************************************** */
/****************************************************

  Calculate a vignetting array for JEM-X based on a
  simple geometry with mask and collimator and centered
  on the FOV.

  In order to avoid problems with freeing memory it
  expects that the necessary memory for the returned
  array has already been allocated when called.

  The array is arranged in a one-dimensional array with
  x varying fastest:

     array(i,j) = array[index]

     index = j * dimx + i

  Version 1.0

  2004-07-30,   Niels J. Westergaard
                Danish Space Research Institute

****************************************************/

int j_vign_matrix( int dimx,          /* Dimension in X direction */
                   int dimy,          /* Dimension in Y direction */
                   float cdelt1,      /* CDELT1 for X in degrees  */
                   float cdelt2,      /* CDELT2 for Y in degrees  */
                   double roll,        /* Roll angle in degrees    */
                   double *vign_matrix,/* Resulting matrix         */
/*                   int chatter,        Verbosity level          */
                   int status ) {

   int indx = 0;
   int indy = 0;
   int dimx_half = 0;
   int dimy_half = 0;

   long index = 0;   /* running index */
   long nelements = 0; /* Size of returned array */

   float cr = 0.0;
   float sr = 0.0;
   float y1 = 0.0;
   float y2 = 0.0;
   float z1 = 0.0;
   float z2 = 0.0;

   if( status != ISDC_OK ) return( status );

   dimx_half = dimx / 2;
   dimy_half = dimy / 2;

   cr = cos( (double) roll * M_PI / 180 );
   sr = sin( (double) roll * M_PI / 180 );

   nelements = ( (long) dimx ) * ( (long) dimy );

   for( index = 0; index < nelements; index++ ) {

      indy = index / dimx;
      indx = index % dimx;
      /* Get position in arcmin - required for function thruput_instr */
      y1 = cdelt1 * ( indx - dimx_half ) * 60.0;
      z1 = cdelt2 * ( indy - dimy_half ) * 60.0;
      y2 = cr * y1 - sr * z1;
      z2 = sr * y1 + cr * z1;

      *(vign_matrix+index) = thruput_instr( y2, z2 );

   }

   return( status );

}


/********************************************

  A function to return the throughput of a JEMX
  mask-collimator combination given the ray input
  angles in Y and Z (unit arcmin!)
  It is equivalent to the IDL function: thruput_instr.pro

  2004-07-30/NJW

*********************************************/

float thruput_instr( float yp, float zp ) {

   float alpha = 0.;
   float swap = 0.;
   float omega = 0.;
   float d = 0.;
   float f = 0.;
   float q = 0.;
   float t = 0.;

   if( yp < 0.0 ) yp = -yp;
   if( zp < 0.0 ) zp = -zp;

   alpha = sqrt(yp*yp + zp*zp) / 3437.75; /* convert to radians */

   if( zp < yp ) {
      swap = zp;
      zp = yp;
      yp = swap;
   }

   omega = atan2( yp, zp );

   /* Area of the two circles */

   d = tan(alpha) * 3401.;  /* Det-mask distance */
   f = circ_area1( d, 267.5, 125. ) / (3.14159265 * 125.0 * 125.0);

   /* What comes through collimator */

   q = (57.0 / 6.6 ) * tan( alpha );

   t = (1.0 - q * cos(omega)) * (1. - q * sin(omega));

   return( f * t );

}


/************************************************

   A function to calculate the overlap area between two circles

   d is the distance between the two centers
   R is the radius of the (larger) circle
   r is the radius of the (smaller) circle

   xR is the distance from the center of the larger circle
      to the line between the two intersection points of the
      two circles

   xr is equivalent to xR but referring to the smaller circle

   aR is the area of the larger circle inside the smaller circle
   ar is the area of the smaller circle inside the larger circle

   2004-07-30/NJW

   Copied from the IDL function: circ_area1.pro

****************************************************/

float circ_area1( float d, float R, float r ) {

   float xR = 0.0;
   float xr = 0.0;
   float aR = 0.0;
   float ar = 0.0;

   float t1 = 0.0;
   float t2 = 0.0;
   float t3 = 0.0;
   float u1 = 0.0;
   float u2 = 0.0;
   float u3 = 0.0;

   if( d >= R + r ) return( 0.0 );  /* No intersection */
   if( d < R - r ) return( PI * r * r ); /* smaller completely inside larger */

   xR = (R*R + d*d - r*r) / (2*d);
   t1 = PI * R * R / 2;
   t2 = R * R * asin( xR / R );
   t3 = xR * sqrt( R*R - xR*xR );

   xr = d - xR;
   u1 = PI * r * r / 2;
   u2 = r * r * asin( xr / r );
   u3 = xr * sqrt( r*r - xr*xr );

   aR = t1 - t2 - t3;
   ar = u1 - u2 - u3;

   return( ar + aR );

}
/* ********************************************************** */
